"""Service for exchanging data with WolkGateway."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import List
from typing import Optional

from wolk_gateway_module.model.message import Message


class ConnectivityService(ABC):
    """Responsible for exchanging data with WolkGateway."""

    @abstractmethod
    def set_inbound_message_listener(
        self, on_inbound_message: Callable[[Message], None]
    ) -> None:
        """
        Set the callback function to handle inbound messages.

        :param on_inbound_message: Callable that handles inbound messages
        :type on_inbound_message: Callable[[Message], None]
        """
        pass

    @abstractmethod
    def set_lastwill_message(self, message: Message) -> None:
        """
        Send offline state for module devices on disconnect.

        :param message: Message to be published
        :type message: Message
        """
        pass

    @abstractmethod
    def add_subscription_topics(self, topics: List[str]) -> None:
        """
        Add subscription topics.

        :param topics: List of topics
        :type topics: List[str]
        """
        pass

    @abstractmethod
    def remove_topics_for_device(self, device_key: str) -> None:
        """
        Remove topics for device from subscription topics.

        :param device_key: Device identifier
        :type device_key: str
        """
        pass

    @abstractmethod
    def connected(self) -> bool:
        """
        Return if currently connected.

        :returns: connected
        :rtype: bool
        """
        pass

    @abstractmethod
    def connect(self) -> Optional[bool]:
        """Establish connection with WolkGateway."""
        pass

    @abstractmethod
    def reconnect(self) -> Optional[bool]:
        """Reestablish connection with WolkGateway."""
        pass

    @abstractmethod
    def disconnect(self) -> Optional[bool]:
        """Terminate connection with WolkGateway."""
        pass

    @abstractmethod
    def publish(self, message: Message) -> bool:
        """
        Publish serialized data to WolkGateway.

        :param message: Message to be published
        :type message: Message
        :returns: result
        :rtype: bool
        """
        pass



"""Device statuses as defined on WolkAbout IoT Platform."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from enum import Enum
from enum import unique


@unique
class DeviceStatus(Enum):
    """
    Enumeration of available device statuses.

    :ivar CONNECTED: Device currently connected
    :vartype CONNECTED: str
    :ivar OFFLINE: Device currently offline
    :vartype OFFLINE: str
    :ivar SERVICE_MODE: Device currently in service mode
    :vartype SERVICE_MODE: str
    :ivar SLEEP: Device currently in sleep mode
    :vartype SLEEP: str
    """

    CONNECTED = "CONNECTED"
    OFFLINE = "OFFLINE"
    SLEEP = "SLEEP"
    SERVICE_MODE = "SERVICE_MODE"



"""Service for exchanging data with WolkGateway."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from time import sleep
from time import time
from typing import Callable
from typing import List
from typing import Optional

from paho.mqtt import client as mqtt

from wolk_gateway_module.connectivity.connectivity_service import (
    ConnectivityService,
)
from wolk_gateway_module.logger_factory import logger_factory
from wolk_gateway_module.model.message import Message


class MQTTConnectivityService(ConnectivityService):
    """Responsible for exchanging data with WolkGateway through MQTT."""

    def __repr__(self) -> str:
        """
        Make string representation of MQTTConnectivityService.

        :returns: representation
        :rtype: str
        """
        return (
            f"MQTTConnectivityService(host='{self.host}', "
            f"port='{self.port}', "
            f"client_id='{self.client_id}', "
            f"topics='{self.topics}', "
            f"qos='{self.qos}', "
            f"lastwill_message='{self.lastwill_message}', "
            f"inbound_message_listener='{self.inbound_message_listener}', "
            f"connected='{self._connected}')"
        )

    def __init__(
        self,
        host: str,
        port: int,
        client_id: str,
        qos: int,
        lastwill_message: Message,
        topics: list,
    ) -> None:
        """
        Prepare MQTT connectivity service for connecting to WolkGateway.

        :param host: Address of WolkGateway
        :type host: str
        :param port: TCP/IP port of WolkGateway used for MQTT connection
        :type port: int
        :param client_id: Unique module identifier
        :type client_id: str
        :param qos: MQTT Quality of Service level (0, 1, 2)
        :type qos: int
        :param lastwill_message: Last will message
        :type lastwill_message: Message
        :param topics: List of topics to subscribe to
        :type topics: list
        """
        self.log = logger_factory.get_logger(str(self.__class__.__name__))

        self.host = host
        self.port = port
        self.client_id = client_id
        self.topics = topics
        self.qos = qos
        self.lastwill_message = lastwill_message
        self.inbound_message_listener: Optional[
            Callable[[Message], None]
        ] = None
        self._connected = False
        self.connected_rc: Optional[int] = None

        self.log.debug(self.__repr__())

    def set_inbound_message_listener(
        self, on_inbound_message: Callable[[Message], None]
    ) -> None:
        """
        Set the callback function ot handle inbound messages.

        :param on_inbound_message: Callable that handles inbound messages
        :type on_inbound_message: Callable[Message]
        """
        self.log.debug(f"Set inbound message listener to {on_inbound_message}")
        self.inbound_message_listener = on_inbound_message

    def set_lastwill_message(self, message: Message) -> None:
        """Send offline state for module devices on disconnect."""
        self.log.debug(f"Set lastwill message: {message}")
        if self._connected:
            self.log.debug("Reconnecting to set lastwill")
            self.disconnect()
            self.lastwill_message = message
            self.connect()
        else:
            self.lastwill_message = message

    def add_subscription_topics(self, topics: List[str]) -> None:
        """
        Add subscription topics.

        :param topics: List of topics
        :type topics: List[str]
        """
        self.log.debug(f"Adding {topics} to {self.topics}")
        self.topics.extend(topics)

    def remove_topics_for_device(self, device_key: str) -> None:
        """
        Remove topics for device from subscription topics.

        :param device_key: Device identifier
        :type device_key: str
        """
        self.log.debug(f"Removing topics for device {device_key}")
        if len(self.topics) == 0:
            return

        for topic in self.topics:
            if device_key in topic:
                self.topics.remove(topic)

    def connected(self) -> bool:
        """
        Return if currently connected.

        :returns: connected
        :rtype: bool
        """
        return self._connected

    def connect(self) -> bool:
        """
        Establish connection with WolkGateway.

        :returns: result
        :rtype: bool

        :raises RuntimeError: Reason for connection being refused
        """
        if self._connected:
            self.log.debug("Already connected")
            return True

        self.client = mqtt.Client(client_id=self.client_id)
        self.client.on_connect = self._on_mqtt_connect
        self.client.on_disconnect = self._on_mqtt_disconnect
        self.client.on_message = self._on_mqtt_message
        self.client.username_pw_set(self.client_id)
        self.client.will_set(
            self.lastwill_message.topic, self.lastwill_message.payload
        )
        self.client.connect(self.host, self.port)
        self.client.loop_start()

        self.log.info(f"Connecting to {self.host}:{self.port} ...")

        timeout = round(time()) + 5

        while True:

            if round(time()) > timeout:
                raise RuntimeError("Connection timed out!")

            if self.connected_rc is None:
                sleep(0.1)
                continue

            if self.connected_rc == 0:
                self._connected = True
                break

            elif self.connected_rc == 1:
                raise RuntimeError(
                    "Connection refused - incorrect protocol version"
                )

            elif self.connected_rc == 2:
                raise RuntimeError(
                    "Connection refused - invalid client identifier"
                )

            elif self.connected_rc == 3:
                raise RuntimeError("Connection refused - server unavailable")

            elif self.connected_rc == 4:
                raise RuntimeError(
                    "Connection refused - bad username or password"
                )

            elif self.connected_rc == 5:
                raise RuntimeError("Connection refused - not authorised")

        self.log.debug(f"Subscribing to topics: {self.topics}")
        for topic in self.topics:
            self.client.subscribe(topic, 2)

        return self._connected

    def reconnect(self) -> bool:
        """
        Terminate existing and create new connection with WolkGateway.

        :returns: result
        :rtype: bool

        :raises RuntimeError: Reason for connection being refused
        """
        self.log.debug("Attempting reconnect")
        self._connected = False
        try:
            self.client.loop_stop()
            self.client.disconnect()
        except AttributeError:
            pass
        try:
            return self.connect()
        except RuntimeError as exception:
            raise exception

    def disconnect(self) -> None:
        """Terminate connection with WolkGateway."""
        self.log.debug(f"Disconnecting from {self.host} : {self.port}")
        try:
            self.client.publish(
                self.lastwill_message.topic, self.lastwill_message.payload
            )
            self.client.loop_stop()
            self.client.disconnect()
        except AttributeError:
            pass
        self._connected = False

    def publish(self, message: Message) -> bool:
        """
        Publish serialized data to WolkGateway.

        :param message: Message to be published
        :type message: Message
        :returns: result
        :rtype: bool
        """
        if not self._connected:
            self.log.warning(f"Not connected, unable to publish {message}")
            return False

        info = self.client.publish(message.topic, message.payload, self.qos)

        if info.rc == mqtt.MQTT_ERR_SUCCESS:
            self.log.debug(f"Published {message}")
            return True
        else:
            return info.is_published()

    def _on_mqtt_message(
        self, client: mqtt.Client, userdata: str, message: mqtt.MQTTMessage
    ) -> None:
        """
        Parse inbound messages and pass them to message listener.

        :param client: Client that received the message
        :type client: paho.mqtt.Client
        :param userdata: Private user data set in Client()
        :type userdata: str
        :param message: Class with members: topic, payload, qos, retain.
        :type message: paho.mqtt.MQTTMessage
        """
        if self.inbound_message_listener is not None:
            self.inbound_message_listener(
                Message(message.topic, message.payload)
            )
        else:
            raise RuntimeError("No inbound message listener is set!")

    def _on_mqtt_connect(
        self, client: mqtt.Client, userdata: str, flags: int, rc: int
    ) -> None:
        """
        Handle when the client receives a CONNACK response from the server.

        :param client: Client that received the message
        :type client: paho.mqtt.Client
        :param userdata: private user data set in Client()
        :type userdata: str
        :param flags: Response flags sent by the broker
        :type flags: int
        :param rc: Connection result
        :type rc: int
        """
        self.log.debug(f"CONNACK: {rc}")
        if rc == 0:  # Connection successful
            self._connected = True
            self.connected_rc = 0
            # Subscribing in on_mqtt_connect() means if we lose the connection
            # and reconnect then subscriptions will be renewed.
            if self.topics:
                for topic in self.topics:
                    self.client.subscribe(topic, 2)
        elif rc == 1:  # Connection refused - incorrect protocol version
            self.connected_rc = 1
        elif rc == 2:  # Connection refused - invalid client identifier
            self.connected_rc = 2
        elif rc == 3:  # Connection refused - server unavailable
            self.connected_rc = 3
        elif rc == 4:  # Connection refused - bad username or password
            self.connected_rc = 4
        elif rc == 5:  # Connection refused - not authorised
            self.connected_rc = 5

    def _on_mqtt_disconnect(
        self, client: mqtt.Client, userdata: str, rc: int
    ) -> None:
        """
        Handle when the client disconnects from the broker.

        :param client: Client that received the message
        :type client: paho.mqtt.Client
        :param userdata: private user data set in Client()
        :type userdata: str
        :param rc: Disconnection result
        :type rc: int
        :raises RuntimeError: Unexpected disconnection
        """
        self.log.debug(f"Disconnect return code: {rc}")
        if rc != 0:
            raise RuntimeError("Unexpected disconnection.")
        self._connected = False
        self.connected_rc = None



"""Handling of messages related to device firmware update."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import List

from wolk_gateway_module.model.firmware_update_status import (
    FirmwareUpdateStatus,
)
from wolk_gateway_module.model.message import Message


class FirmwareUpdateProtocol(ABC):
    """Parse inbound messages and serialize outbound firmware messages."""

    @abstractmethod
    def get_inbound_topics_for_device(self, device_key: str) -> List[str]:
        """
        Return list of inbound topics for given device key.

        :param device_key: Device key for which to create topics
        :type device_key: str

        :returns: inbound_topics
        :rtype: list
        """
        pass

    @abstractmethod
    def make_update_message(
        self, device_key: str, status: FirmwareUpdateStatus
    ) -> Message:
        """
        Make message from device firmware update status.

        :param device_key: Device key to which the firmware update belongs to
        :type device_key: str
        :param status: Device firmware update status
        :type status: FirmwareUpdateStatus

        :returns: message
        :rtype: Message
        """
        pass

    @abstractmethod
    def make_version_message(
        self, device_key: str, firmware_verison: str
    ) -> Message:
        """
        Make message from device firmware update version.

        :param device_key: Device key to which the firmware update belongs to
        :type device_key: str
        :param firmware_verison: Current firmware version
        :type firmware_verison: str

        :returns: message
        :rtype: Message
        """
        pass

    @abstractmethod
    def is_firmware_install_command(self, message: Message) -> bool:
        """
        Check if received message is firmware install command.

        :param message: Message received
        :type message: Message

        :returns: is_firmware_install_command
        :rtype: bool
        """
        pass

    @abstractmethod
    def is_firmware_abort_command(self, message: Message) -> bool:
        """
        Check if received message is firmware abort command.

        :param message: Message received
        :type message: Message

        :returns: is_firmware_abort_command
        :rtype: bool
        """
        pass

    @abstractmethod
    def make_firmware_file_path(self, message: Message) -> str:
        """
        Extract file path from firmware install message.

        :param message: Message received
        :type message: Message

        :returns: firmware_file_path
        :rtype: str
        """
        pass

    @abstractmethod
    def extract_key_from_message(self, message: Message) -> str:
        """
        Return device key from message.

        :param message: Message received
        :type message: Message

        :returns: device_key
        :rtype: str
        """
        pass



"""Tests for JsonFirmwareUpdateProtocol."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
import sys
import unittest

sys.path.append("..")  # noqa


from wolk_gateway_module.json_firmware_update_protocol import (
    JsonFirmwareUpdateProtocol,
)
from wolk_gateway_module.model.firmware_update_status import (
    FirmwareUpdateStatus,
    FirmwareUpdateState,
    FirmwareUpdateErrorCode,
)
from wolk_gateway_module.model.message import Message


class JsonFirmwareUpdateProtocolTests(unittest.TestCase):
    """JsonFirmwareUpdateProtocol Tests."""

    DEVICE_PATH_PREFIX = "d/"
    FIRMWARE_UPDATE_INSTALL_TOPIC_ROOT = "p2d/firmware_update_install/"
    FIRMWARE_UPDATE_ABORT_TOPIC_ROOT = "p2d/firmware_update_abort/"
    FIRMWARE_UPDATE_STATUS_TOPIC_ROOT = "d2p/firmware_update_status/"
    FIRMWARE_VERSION_UPDATE_TOPIC_ROOT = "d2p/firmware_version_update/"

    def test_get_inbound_topics_for_device(self):
        """Test that returned list is correct for given device key."""
        json_firmware_update_protocol = JsonFirmwareUpdateProtocol()
        device_key = "some_key"

        expected = [
            self.FIRMWARE_UPDATE_INSTALL_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
            self.FIRMWARE_UPDATE_ABORT_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
        ]

        self.assertEqual(
            expected,
            json_firmware_update_protocol.get_inbound_topics_for_device(
                device_key
            ),
        )

    def test_make_firmware_update_installation_status_message(self):
        """Test that firmware update status message is created correctly."""
        json_firmware_update_protocol = JsonFirmwareUpdateProtocol()
        device_key = "some_key"
        status = FirmwareUpdateStatus(FirmwareUpdateState.INSTALLATION)

        expected = Message(
            self.FIRMWARE_UPDATE_STATUS_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
            json.dumps({"status": "INSTALLATION"}),
        )

        self.assertEqual(
            expected,
            json_firmware_update_protocol.make_update_message(
                device_key, status
            ),
        )

    def test_make_firmware_update_error_file_system_status_message(self):
        """Test that firmware update status message is created correctly."""
        json_firmware_update_protocol = JsonFirmwareUpdateProtocol()
        device_key = "some_key"
        status = FirmwareUpdateStatus(
            FirmwareUpdateState.INSTALLATION,
            FirmwareUpdateErrorCode.FILE_SYSTEM_ERROR,
        )

        expected = Message(
            self.FIRMWARE_UPDATE_STATUS_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
            json.dumps(
                {
                    "status": status.status.value,
                    "error": status.error_code.value,
                }
            ),
        )

        self.assertEqual(
            expected,
            json_firmware_update_protocol.make_update_message(
                device_key, status
            ),
        )

    def test_make_version_message(self):
        """Test that firmware version message is created correctly."""
        json_firmware_update_protocol = JsonFirmwareUpdateProtocol()
        device_key = "some_key"
        firmware_version = "v1.0"

        expected = Message(
            self.FIRMWARE_VERSION_UPDATE_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key,
            firmware_version,
        )

        self.assertEqual(
            expected,
            json_firmware_update_protocol.make_version_message(
                device_key, firmware_version
            ),
        )

    def test_is_firmware_install_command(self):
        """Test that message is firmware install command."""
        json_firmware_update_protocol = JsonFirmwareUpdateProtocol()

        message = Message(self.FIRMWARE_UPDATE_INSTALL_TOPIC_ROOT)

        self.assertTrue(
            json_firmware_update_protocol.is_firmware_install_command(message)
        )

    def test_is_firmware_abort_command(self):
        """Test that message is firmware abort command."""
        json_firmware_update_protocol = JsonFirmwareUpdateProtocol()
        message = Message(self.FIRMWARE_UPDATE_ABORT_TOPIC_ROOT)

        self.assertTrue(
            json_firmware_update_protocol.is_firmware_abort_command(message)
        )

    def test_make_firmware_file_path(self):
        """Test firmare file path is extracted correctly."""
        json_firmware_update_protocol = JsonFirmwareUpdateProtocol()
        expected = "some/path/to/file"

        message = Message(
            self.FIRMWARE_UPDATE_INSTALL_TOPIC_ROOT,
            json.dumps({"fileName": expected}),
        )

        self.assertEqual(
            expected,
            json_firmware_update_protocol.make_firmware_file_path(message),
        )

    def test_extract_key_from_message(self):
        """Test that device key is correctly extracted from abort message."""
        json_firmware_update_protocol = JsonFirmwareUpdateProtocol()
        device_key = "some_key"

        message = Message(
            self.FIRMWARE_UPDATE_ABORT_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key
        )

        self.assertEqual(
            device_key,
            json_firmware_update_protocol.extract_key_from_message(message),
        )


if __name__ == "__main__":
    unittest.main()



"""MQTT message model."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from typing import Optional
from typing import Union


@dataclass
class Message:
    """
    MQTT message identified by topic and payload.

    :ivar topic: Topic where the message is from or will be sent to
    :vartype topic: str
    :ivar payload: Content of the message
    :vartype payload: bytes or str or bytearray or None
    """

    topic: str
    payload: Optional[Union[str, bytes, bytearray]] = field(default=None)



"""Available reading type measurement units."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from enum import Enum


class ReadingTypeMeasurementUnit(Enum):
    """Enumeration of defined reading type measurement units."""

    BIT = "bit"
    PERCENT = "%"
    CO2_MOL = "mol"
    X10C = "X10℃"
    X100V = "X100V"
    X10PA = "X10Pa"
    NUMERIC = ""
    KELVIN = "K"
    CELSIUS = "℃"
    FAHRENHEIT = "°F"
    CELSIUS_X2 = "X2C"
    CELSIUS_X10 = "X10℃"
    PASCAL = "Pa"
    MILLIMETER_OF_MERCURY = "mmHg"
    INCH_OF_MERCURY = "inHg"
    BAR = "bar"
    ATMOSPHERE = "atm"
    MILLIBAR = "mb"
    PRESSURE_PERCENT = "%"
    MILLIBAR_X10 = "X10mb"
    MICROBAR = "μbar"
    KILO_PASCAL = "kPa"
    HUMIDITY_PERCENT = "%"
    HUMIDITY_PERCENT_X10 = "X10%"
    VOLT = "V"
    MILLIVOLT = "mV"
    CENTIVOLT = "cV"
    MOVEMENT = ""
    LIGHT_PERCENT = "%"
    METRES_PER_SQUARE_SECOND = "m/s²"
    GRAVITY = "g"
    GYROSCOPE = "deg/s"
    LOCATION = ""
    BEATS_PER_MINUTE = "bpm"
    BATTERY = "Wh"
    BATTERY_X1000 = "kWh"
    BREATHS_PER_MINUTE = "bpm"
    CALORIES = "cal"
    AMPERE = "A"
    MILLIAMPERE = "mA"
    WATT = "W"
    HORSEPOWER = "hp"
    MILLIWATT = "mW"
    METER = ""
    OUNCE_LIQUID_US = "oz"
    GALLON_UK = "gallon_u"
    GALLON_DRY_US = "gallon_d"
    LITRE = "L"
    OUNCE_LIQUID_UK = "oz_uk"
    MILLILITRE = "mL"
    METRE = "m"
    MILE = "mi"
    POINT = "pt"
    FOOT = "ft"
    INCH = "in"
    PARSEC = "pc"
    YARD = "yd"
    MILLIMETER = "mm"
    CENTIMETER = "cm"
    KILOMETER = "km"
    KILOGRAM = "kg"
    GALLON_LIQUID_US = "gal"
    ATOMIC_MASS = "u"
    TON_UK = "ton_uk"
    METRIC_TON = "t"
    TON_US = "ton_us"
    ELECTRON_MASS = "me"
    POUND = "lb"
    OUNCE = "oz"
    MILLIGRAM = "mg"
    GRAM = "g"
    DECIBEL = "dB"
    KNOT = "kn"
    KILOMETERS_PER_HOUR = "km/h"
    MILES_PER_HOUR = "mph"
    MACH = "Mach"
    SPEED_OF_LIGHT = "c"
    METER_PER_SECOND = "m/s"
    SECOND = "s"
    MINUTE = "min"
    HOUR = "h"
    MONTH = "month"
    DAY = "day"
    WEEK = "week"
    YEAR = "year"
    TESLA = "T"
    GAUSS = "G"
    MICRO_TESLA = "μT"
    SIEVERT = "C"
    BECQUEREL = "Bq"
    RUTHERFORD = "Rd"
    ROENTGEN = "Roentgen"
    RADIATION_DOSE_EFFECTIVE = "rem"
    CURIE = "Ci"
    NEWTON = "N"
    POUND_FORCE = "lbf"
    GRAVITY_FORCE = "grav"
    KILOGRAM_FORCE = "kgf"
    DYNE = "dyn"
    SQUARE_METRE = "m²"
    CUBIC_METRE = "m³"
    BYTE = "byte"
    GRADE = "grade"
    HECTARE = "ha"
    CUBIC_INCH = "in³"
    REVOLUTION = "rev"
    CENTIRADIAN = "centirad"
    RAD = "rd"
    COMPUTER_POINT = "pixel"
    DEGREE_ANGLE = "°"
    SECOND_ANGLE = "″"
    MINUTE_ANGLE = "′"
    SPHERE = "sphere"
    ARE = "a"
    RADIAN = "rad"
    HERTZ = "Hz"
    MEGA_HERTZ = "MHz"
    GIGA_HERTZ = "GHz"
    WEBER = "Wb"
    MAXWELL = "Mx"
    FARAD = "F"
    FARADAY = "Fd"
    OHM = "Ω"
    FRANKLIN = "Fr"
    ELECTRIC_CHARGE = "e"
    GILBERT = "Gi"
    JOULE = "J"
    ELECTRON_VOLT = "eV"
    HENRY = "H"
    SIEMENS = "S"
    LUMEN = "lm"
    CANDELA = "cd"
    LUX = "lx"
    TEXT = ""
    BOOLEAN = ""



"""Stub method for providing current device actuator status."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Tuple
from typing import Union

from wolk_gateway_module.model.actuator_state import ActuatorState


def get_actuator_status(
    device_key: str, reference: str
) -> Tuple[ActuatorState, Union[bool, int, float, str]]:
    """
    Get current actuator status identified by device key and reference.

    Reads the status of actuator from the device
    and returns it as a tuple containing the actuator state and current value.

    Must be implemented as non blocking.
    Must be implemented as thread safe.

    :param device_key: Device key to which the actuator belongs to
    :type device_key: str
    :param reference: Actuator reference
    :type reference: str
    :returns: (state, value)
    :rtype: (ActuatorState, bool or int or float or str)
    """
    pass



"""Handling of device registration using JSON_PROTOCOL."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json

from wolk_gateway_module.logger_factory import logger_factory
from wolk_gateway_module.model.device_registration_request import (
    DeviceRegistrationRequest,
)
from wolk_gateway_module.model.device_registration_response import (
    DeviceRegistrationResponse,
)
from wolk_gateway_module.model.device_registration_response_result import (
    DeviceRegistrationResponseResult,
)
from wolk_gateway_module.model.message import Message
from wolk_gateway_module.protocol.registration_protocol import (
    RegistrationProtocol,
)


class JsonRegistrationProtocol(RegistrationProtocol):
    """Send device registration requests and handle their responses."""

    DEVICE_PATH_PREFIX = "d/"
    CHANNEL_WILDCARD = "#"
    DEVICE_REGISTRATION_REQUEST_TOPIC_ROOT = "d2p/register_subdevice_request/"
    DEVICE_REGISTRATION_RESPONSE_TOPIC_ROOT = (
        "p2d/register_subdevice_response/"
    )

    def __init__(self) -> None:
        """Create object."""
        self.log = logger_factory.get_logger(str(self.__class__.__name__))

    def __repr__(self) -> str:
        """
        Make string representation of JsonRegistrationProtocol.

        :returns: representation
        :rtype: str
        """
        return "JsonRegistrationProtocol()"

    def get_inbound_topics_for_device(self, device_key: str) -> list:
        """
        Return list of inbound topics for given device key.

        :param device_key: Device key for which to create topics
        :type device_key: str

        :returns: inbound_topics
        :rtype: list
        """
        inbound_topics = [
            self.DEVICE_REGISTRATION_RESPONSE_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + device_key
        ]
        self.log.debug(f"Inbound topics for {device_key} : {inbound_topics}")

        return inbound_topics

    def extract_key_from_message(self, message: Message) -> str:
        """
        Return device key from message.

        :param message: Message received
        :type message: Message

        :returns: device_key
        :rtype: str
        """
        device_key = message.topic.split("/")[-1]
        self.log.debug(f"Made {device_key} from {message}")

        return device_key

    def is_registration_response_message(self, message: Message) -> bool:
        """
        Check if message is device registration response.

        :param message: Message received
        :type message: Message

        :returns: is_device_registration_response
        :rtype: bool
        """
        is_device_registration_response = message.topic.startswith(
            self.DEVICE_REGISTRATION_RESPONSE_TOPIC_ROOT
        )
        self.log.debug(
            f"Is {message} device registration response "
            f"message: {is_device_registration_response}"
        )

        return is_device_registration_response

    def make_registration_message(
        self, request: DeviceRegistrationRequest
    ) -> Message:
        """
        Make message from device registration request.

        :param request: Device registration request
        :type request: DeviceRegistrationRequest

        :returns: message
        :rtype: Message
        """
        request_dict = {
            "name": request.name,
            "deviceKey": request.key,
            "defaultBinding": request.default_binding,
            "typeParameters": request.template.type_parameters,
            "connectivityParameters": request.template.connectivity_parameters,
        }

        if request.template.supports_firmware_update:
            request_dict["firmwareUpdateType"] = "DFU"
        else:
            request_dict["firmwareUpdateType"] = ""

        request_dict["sensors"] = [
            sensor.to_dto() for sensor in request.template.sensors
        ]

        request_dict["actuators"] = [
            actuator.to_dto() for actuator in request.template.actuators
        ]

        request_dict["alarms"] = [
            alarm.to_dto() for alarm in request.template.alarms
        ]

        request_dict["configurations"] = [
            configuration.to_dto()
            for configuration in request.template.configurations
        ]

        if (
            "supportsFirmwareUpdate"
            not in request.template.firmware_update_parameters
        ):
            request.template.firmware_update_parameters[
                "supportsFirmwareUpdate"
            ] = request.template.supports_firmware_update

        request_dict[
            "firmwareUpdateParameters"
        ] = request.template.firmware_update_parameters

        message = Message(
            self.DEVICE_REGISTRATION_REQUEST_TOPIC_ROOT
            + self.DEVICE_PATH_PREFIX
            + request.key,
            json.dumps(request_dict, ensure_ascii=False)
            .encode("utf-8")
            .decode("utf-8"),
        )
        self.log.debug(f"Made {message} from {request}")

        return message

    def make_registration_response(
        self, message: Message
    ) -> DeviceRegistrationResponse:
        """
        Make device registration response from message.

        :param message: Message received
        :rtype message: Message

        :returns: device_registration_response
        :rtype: DeviceRegistrationResponse
        """
        response = json.loads(message.payload)  # type: ignore

        result = DeviceRegistrationResponseResult.ERROR_UNKNOWN

        for enumeration in DeviceRegistrationResponseResult:
            if response["result"] == enumeration.value:
                result = enumeration
                break

        description = (
            response["description"] if "description" in response else ""
        )

        device_registration_response = DeviceRegistrationResponse(
            response["payload"]["deviceKey"], result, description
        )
        self.log.debug(f"Made {device_registration_response} from {message}")

        return device_registration_response



"""Configuration command received from WolkAbout IoT Platform."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from enum import auto
from enum import Enum
from enum import unique
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

Configuration = Dict[
    str,
    Union[
        int,
        float,
        bool,
        str,
        Tuple[int, int],
        Tuple[int, int, int],
        Tuple[float, float],
        Tuple[float, float, float],
        Tuple[str, str],
        Tuple[str, str, str],
    ],
]


@unique
class ConfigurationCommandType(Enum):
    """
    Configuration command type.

    :ivar GET: Get current configuration options
    :vartype GET: int
    :ivar SET: Set configuration to value
    :vartype SET: int
    """

    GET = auto()
    SET = auto()


@dataclass
class ConfigurationCommand:
    """
    Configuration command with command and optionally value.

    :ivar command: Configuration command received
    :vartype command: int
    :ivar value: Set configuration to value
    :vartype value: Optional[dict]
    """

    command: ConfigurationCommandType
    value: Optional[Configuration] = field(default=None)



"""Device registration request model."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field

from wolk_gateway_module.model.device_template import DeviceTemplate


@dataclass
class DeviceRegistrationRequest:
    """
    Request for device registration.

    :ivar name: Device name
    :vartype name: str
    :ivar key: Unique device key
    :vartype key: str
    :ivar template: Device template
    :vartype template: DeviceTemplate
    :ivar default_binding: Create semantic group for device on Platform
    :vartype default_binding: bool
    """

    name: str
    key: str
    template: DeviceTemplate = field(default_factory=DeviceTemplate)
    default_binding: bool = field(default=True)



"""Handling of inbound and outbound messages related to device status."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import List

from wolk_gateway_module.model.device_status import DeviceStatus
from wolk_gateway_module.model.message import Message


class StatusProtocol(ABC):
    """Parse inbound messages and serialize device status messages."""

    @abstractmethod
    def get_inbound_topics_for_device(self, device_key: str) -> List[str]:
        """
        Return list of inbound topics for given device key.

        :param device_key: Device key for which to create topics
        :type device_key: str

        :returns: inbound_topics
        :rtype: list
        """
        pass

    @abstractmethod
    def is_device_status_request_message(self, message: Message) -> bool:
        """
        Check if message is device status request.

        :param message: Message received
        :type message: Message

        :returns: is_device_status_request
        :rtype: bool
        """
        pass

    @abstractmethod
    def make_device_status_response_message(
        self, device_status: DeviceStatus, device_key: str
    ) -> Message:
        """
        Make message from device status response.

        :param device_status: Device's current status
        :type device_status: DeviceStatus
        :param device_key: Device to which the status belongs to
        :type device_key: str

        :returns: message
        :rtype: Message
        """
        pass

    @abstractmethod
    def make_device_status_update_message(
        self, device_status: DeviceStatus, device_key: str
    ) -> Message:
        """
        Make message from device status update.

        :param device_status: Device's current status
        :type device_status: DeviceStatus
        :param device_key: Device to which the status belongs to
        :type device_key: str

        :returns: message
        :rtype: Message
        """
        pass

    @abstractmethod
    def make_last_will_message(self, device_keys: List[str]) -> Message:
        """
        Make last will message from list of device keys.

        :param device_keys: List of device keys
        :type device_keys: list(str)

        :returns: message
        :rtype: Message
        """
        pass

    @abstractmethod
    def extract_key_from_message(self, message: Message) -> str:
        """
        Extract device key from message.

        :param message: Message received
        :type message: Message

        :returns: device_key
        :rtype: str
        """
        pass



"""Sensor reading model."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from typing import Optional
from typing import Tuple
from typing import Union


@dataclass
class SensorReading:
    """
    Holds information about a sensor reading.

    :ivar reference: Device sensor's reference as defined in device template
    :vartype reference: str
    :ivar value: Data that the sensor reading yielded
    :vartype value: bool or int or float or str or Tuple[int, int] or Tuple[int, int, int] or Tuple[float, float] or Tuple[float, float, float] or Tuple[str, str] or Tuple[str, str, str]
    :ivar timestamp:  Unix timestamp in miliseconds. If not provided, Platform will assign timestamp when it receives it.
    :vartype timestamp: Optional[int]
    """

    reference: str
    value: Union[
        bool,
        int,
        float,
        str,
        Tuple[int, int],
        Tuple[int, int, int],
        Tuple[float, float],
        Tuple[float, float, float],
        Tuple[str, str],
        Tuple[str, str, str],
    ]
    timestamp: Optional[int] = field(default=None)



"""Handling of inbound and outbound messages related to device registraiton."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import List

from wolk_gateway_module.model.device_registration_request import (
    DeviceRegistrationRequest,
)
from wolk_gateway_module.model.device_registration_response import (
    DeviceRegistrationResponse,
)
from wolk_gateway_module.model.message import Message


class RegistrationProtocol(ABC):
    """Parse inbound messages and serialize outbound registration messages."""

    @abstractmethod
    def get_inbound_topics_for_device(self, device_key: str) -> List[str]:
        """
        Return list of inbound topics for given device key.

        :param device_key: Device key for which to create topics
        :type device_key: str

        :returns: inbound_topics
        :rtype: list
        """
        pass

    @abstractmethod
    def extract_key_from_message(self, message: Message) -> str:
        """
        Return device key from message.

        :param message: Message received
        :type message: Message

        :returns: device_key
        :rtype: str
        """
        pass

    @abstractmethod
    def is_registration_response_message(self, message: Message) -> bool:
        """
        Check if message is device registration response.

        :param message: Message received
        :type message: Message

        :returns: is_device_registration_response
        :rtype: bool
        """
        pass

    @abstractmethod
    def make_registration_message(
        self, request: DeviceRegistrationRequest
    ) -> Message:
        """
        Make message from device registration request.

        :param request: Device registration request
        :type request: DeviceRegistrationRequest

        :returns: message
        :rtype: Message
        """
        pass

    @abstractmethod
    def make_registration_response(
        self, message: Message
    ) -> DeviceRegistrationResponse:
        """
        Make device registration response from message.

        :param message: Message received
        :rtpe message: Message

        :returns: device_registration_response
        :rtype: DeviceRegistrationResponse
        """
        pass



"""Firmware update status model."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from enum import unique
from typing import Optional


@unique
class FirmwareUpdateState(Enum):
    """
    Enumeration of available firmware update states.

    :ivar ABORTED: Firmware installation aborted
    :vartype ABORTED: str
    :ivar COMPLETED: Firmware installation completed
    :vartype COMPLETED: str
    :ivar ERROR: Firmware installation error
    :vartype ERROR: str
    :ivar INSTALLATION: Firmware installation in progress
    :vartype INSTALLATION: str
    """

    INSTALLATION = "INSTALLATION"
    COMPLETED = "COMPLETED"
    ERROR = "ERROR"
    ABORTED = "ABORTED"


@unique
class FirmwareUpdateErrorCode(Enum):
    """
    Enumeration of possible firmware update errors.

    :ivar DEVICE_NOT_PRESENT: Unable to pass firmware install command to device
    :vartype DEVICE_NOT_PRESENT: int
    :ivar FILE_NOT_PRESENT: Firmware file was not present at specified location
    :vartype FILE_NOT_PRESENT: int
    :ivar FILE_SYSTEM_ERROR: File system error occurred
    :vartype FILE_SYSTEM_ERROR: int
    :ivar INSTALLATION_FAILED: Firmware installation failed
    :vartype INSTALLATION_FAILED: int
    :ivar UNSPECIFIED_ERROR: Unspecified error occurred
    :vartype UNSPECIFIED_ERROR: int
    """

    UNSPECIFIED_ERROR = 0
    FILE_NOT_PRESENT = 1
    FILE_SYSTEM_ERROR = 2
    INSTALLATION_FAILED = 3
    DEVICE_NOT_PRESENT = 4


@dataclass
class FirmwareUpdateStatus:
    """
    Holds information about current firmware update status.

    :ivar status: Firmware update status
    :vartype status: FirmwareUpdateState
    :ivar error_code: Description of error that occured
    :vartype error_code: Optional[FirmwareUpdateErrorCode]
    """

    status: FirmwareUpdateState
    error_code: Optional[FirmwareUpdateErrorCode] = field(default=None)



"""Stub method for providing current device configuration options."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Dict
from typing import Tuple
from typing import Union


def get_configuration(
    device_key: str,
) -> Dict[
    str,
    Union[
        int,
        float,
        bool,
        str,
        Tuple[int, int],
        Tuple[int, int, int],
        Tuple[float, float],
        Tuple[float, float, float],
        Tuple[str, str],
        Tuple[str, str, str],
    ],
]:
    """
    Get current configuration options.

    Reads device configuration and returns it as a dictionary
    with device configuration reference as key,
    and device configuration value as value.
    Must be implemented as non blocking.
    Must be implemented as thread safe.

    :param device_key: Device identifier
    :type device_key: str
    :returns: configuration
    :rtype: dict
    """
    pass



"""Handling of outbound and inbound messages related to device data."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
from typing import List

from wolk_gateway_module.logger_factory import logger_factory
from wolk_gateway_module.model.actuator_command import ActuatorCommand
from wolk_gateway_module.model.actuator_command import ActuatorCommandType
from wolk_gateway_module.model.actuator_status import ActuatorStatus
from wolk_gateway_module.model.alarm import Alarm
from wolk_gateway_module.model.configuration_command import (
    ConfigurationCommand,
)
from wolk_gateway_module.model.configuration_command import (
    ConfigurationCommandType,
)
from wolk_gateway_module.model.message import Message
from wolk_gateway_module.model.sensor_reading import SensorReading
from wolk_gateway_module.protocol.data_protocol import DataProtocol


class JsonDataProtocol(DataProtocol):
    """Parse inbound messages and serialize outbound messages."""

    DEVICE_PATH_PREFIX = "d/"
    REFERENCE_PATH_PREFIX = "r/"
    CHANNEL_WILDCARD = "#"
    CHANNEL_DELIMITER = "/"
    SENSOR_READING = "d2p/sensor_reading/"
    ALARM = "d2p/events/"
    ACTUATOR_SET = "p2d/actuator_set/"
    ACTUATOR_GET = "p2d/actuator_get/"
    ACTUATOR_STATUS = "d2p/actuator_status/"
    CONFIGURATION_SET = "p2d/configuration_set/"
    CONFIGURATION_GET = "p2d/configuration_get/"
    CONFIGURATION_STATUS = "d2p/configuration_get/"

    def __init__(self) -> None:
        """Create object."""
        self.log = logger_factory.get_logger(str(self.__class__.__name__))

    def __repr__(self) -> str:
        """
        Make string representation of JsonDataProtocol.

        :returns: representation
        :rtype: str
        """
        return "JsonDataProtocol()"

    def get_inbound_topics_for_device(self, device_key: str) -> list:
        """
        Return list of inbound topics for given device key.

        :param device_key: Device key for which to create topics
        :type device_key: str

        :returns: inbound_topics
        :rtype: list
        """
        inbound_topics = [
            self.ACTUATOR_SET
            + self.DEVICE_PATH_PREFIX
            + device_key
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + self.CHANNEL_WILDCARD,
            self.ACTUATOR_GET
            + self.DEVICE_PATH_PREFIX
            + device_key
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + self.CHANNEL_WILDCARD,
            self.CONFIGURATION_SET + self.DEVICE_PATH_PREFIX + device_key,
            self.CONFIGURATION_GET + self.DEVICE_PATH_PREFIX + device_key,
        ]
        self.log.debug(f"Inbound topics for {device_key} : {inbound_topics}")

        return inbound_topics

    def is_actuator_get_message(self, message: Message) -> bool:
        """
        Check if message is actuator get command.

        :param message: Message received
        :type message: Message

        :returns: is_actuator_get_message
        :rtype: bool
        """
        is_actuator_get_message = message.topic.startswith(self.ACTUATOR_GET)
        self.log.debug(
            f"Is {message} actuator get message: {is_actuator_get_message}"
        )

        return is_actuator_get_message

    def is_actuator_set_message(self, message: Message) -> bool:
        """
        Check if message is actuator set command.

        :param message: Message received
        :type message: Message

        :returns: is_actuator_set_message
        :rtype: bool
        """
        is_actuator_set_message = message.topic.startswith(self.ACTUATOR_SET)
        self.log.debug(
            f"Is {message} actuator set message: {is_actuator_set_message}"
        )

        return is_actuator_set_message

    def is_configuration_get_message(self, message: Message) -> bool:
        """
        Check if message is configuration get command.

        :param message: Message received
        :type message: Message

        :returns: is_configuration_get_message
        :rtype: bool
        """
        is_configuration_get_message = message.topic.startswith(
            self.CONFIGURATION_GET
        )
        self.log.debug(
            f"Is {message} configuration get "
            f"message: {is_configuration_get_message}"
        )

        return is_configuration_get_message

    def is_configuration_set_message(self, message: Message) -> bool:
        """
        Check if message is configuration set command.

        :param message: Message received
        :type message: Message

        :returns: is_configuration_set_message
        :rtype: bool
        """
        is_configuration_set_message = message.topic.startswith(
            self.CONFIGURATION_SET
        )
        self.log.debug(
            f"Is {message} configuration set "
            f"message: {is_configuration_set_message}"
        )

        return is_configuration_set_message

    def extract_key_from_message(self, message: Message) -> str:
        """
        Extract device key from message.

        :param message: Message received
        :type message: Message

        :returns: device_key
        :rtype: str
        """
        if self.REFERENCE_PATH_PREFIX in message.topic:
            device_key = message.topic.split("/")[-3]
        else:
            device_key = message.topic.split("/")[-1]
        self.log.debug(f"Made {device_key} from {message}")

        return device_key

    def make_actuator_command(self, message: Message) -> ActuatorCommand:
        """
        Make actuator command from message.

        :param message: Message received
        :type message: Message

        :returns: actuator_command
        :rtype: ActuatorCommand
        """
        reference = message.topic.split("/")[-1]
        if self.is_actuator_set_message(message):
            command = ActuatorCommandType.SET
            payload = json.loads(message.payload)  # type: ignore
            value = payload["value"]
            if "\n" in str(value):
                value = value.replace("\n", "\\n")
                value = value.replace("\r", "")

            if "true" == str(value):
                value = True
            elif "false" == str(value):
                value = False

            else:
                try:
                    if any("." in char for char in value):
                        value = float(value)
                    else:
                        value = int(value)
                except (ValueError, TypeError):
                    pass
        elif self.is_actuator_get_message(message):
            command = ActuatorCommandType.GET
            value = None

        actuator_command = ActuatorCommand(reference, command, value)
        self.log.debug(f"Made {actuator_command} from {message}")

        return actuator_command

    def make_configuration_command(
        self, message: Message
    ) -> ConfigurationCommand:
        """
        Make configuration command from message.

        :param message: Message received
        :type message: Message

        :returns: configuration_command
        :rtype: ConfigurationCommand
        """
        if self.is_configuration_set_message(message):
            command = ConfigurationCommandType.SET
            payload = json.loads(message.payload)  # type: ignore
            for reference, value in payload.items():
                if "\n" in str(value):
                    value = value.replace("\n", "\\n")
                    value = value.replace("\r", "")

                if "true" == str(value):
                    payload[reference] = True
                elif "false" == str(value):
                    payload[reference] = False

                else:
                    try:
                        if any("." in char for char in value):
                            payload[reference] = float(value)
                        else:
                            payload[reference] = int(value)
                    except (ValueError, TypeError):
                        pass

                if isinstance(payload[reference], (int, float, bool)):
                    pass
                else:
                    if "," in value:
                        values_list = value.split(",")
                        try:
                            if any("." in value for value in values_list):
                                values_list = [
                                    float(value) for value in values_list
                                ]
                            else:
                                values_list = [
                                    int(value) for value in values_list
                                ]
                        except ValueError:
                            pass
                        payload[reference] = tuple(values_list)
                values = payload

        elif self.is_configuration_get_message(message):
            command = ConfigurationCommandType.GET
            values = None

        configuration_command = ConfigurationCommand(command, values)
        self.log.debug(f"Made {configuration_command} from {message}")

        return configuration_command

    def make_sensor_reading_message(
        self, device_key: str, sensor_reading: SensorReading
    ) -> Message:
        """
        Make message from sensor reading for device key.

        :param device_key: Device on which the sensor reading occurred
        :type device_key: str
        :param sensor_reading: Sensor reading data
        :type sensor_reading: SensorReading

        :returns: message
        :rtype: Message
        """
        topic = (
            self.SENSOR_READING
            + self.DEVICE_PATH_PREFIX
            + device_key
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + sensor_reading.reference
        )

        if isinstance(sensor_reading.value, tuple):
            sensor_reading.value = ",".join(map(str, sensor_reading.value))
        elif isinstance(sensor_reading.value, bool):
            sensor_reading.value = str(sensor_reading.value).lower()

        if sensor_reading.timestamp is not None:
            payload = json.dumps(
                {
                    "data": str(sensor_reading.value),
                    "utc": int(sensor_reading.timestamp),
                }
            )
        else:
            payload = json.dumps({"data": str(sensor_reading.value)})

        message = Message(topic, payload)
        self.log.debug(
            f"Made {message} from {sensor_reading} and {device_key}"
        )

        return message

    def make_sensor_readings_message(
        self,
        device_key: str,
        sensor_readings: List[SensorReading],
        timestamp: int = None,
    ) -> Message:
        """
        Make message from multiple sensor readings for device key.

        :param device_key: Device on which the sensor reading occurred
        :type device_key: str
        :param sensor_reading: Sensor readings data
        :type sensor_reading: List[SensorReading]
        :param timestamp: Timestamp
        :type timestamp: Optional[int]

        :returns: message
        :rtype: Message
        """
        topic = self.SENSOR_READING + self.DEVICE_PATH_PREFIX + device_key

        payload = {}
        for sensor_reading in sensor_readings:
            if isinstance(sensor_reading.value, tuple):
                sensor_reading.value = ",".join(map(str, sensor_reading.value))
            elif isinstance(sensor_reading.value, bool):
                sensor_reading.value = str(sensor_reading.value).lower()

            payload[sensor_reading.reference] = sensor_reading.value

        if timestamp is not None:
            payload["utc"] = timestamp

        message = Message(topic, json.dumps(payload))
        self.log.debug(
            f"Made {message} from {sensor_readings} and {device_key} "
            f"and timestamp {timestamp}"
        )

        return message

    def make_alarm_message(self, device_key: str, alarm: Alarm) -> Message:
        """
        Make message from alarm for device key.

        :param device_key: Device on which the alarm occurred
        :type device_key: str
        :param alarm: Alarm data
        :type alarm: Alarm

        :returns: message
        :rtype: Message
        """
        topic = (
            self.ALARM
            + self.DEVICE_PATH_PREFIX
            + device_key
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + alarm.reference
        )

        if alarm.timestamp is not None:
            payload = json.dumps(
                {
                    "data": str(alarm.active).lower(),
                    "utc": int(alarm.timestamp),
                }
            )
        else:
            payload = json.dumps({"data": str(alarm.active).lower()})

        message = Message(topic, payload)
        self.log.debug(f"Made {message} from {alarm} and {device_key}")

        return message

    def make_actuator_status_message(
        self, device_key: str, actuator_status: ActuatorStatus
    ) -> Message:
        """
        Make message from actuator status for device key.

        :param device_key: Device on which the actuator status occurred
        :type device_key: str
        :param actuator_status: Actuator status data
        :type actuator_status: ActuatorStatus

        :returns: message
        :rtype: Message
        """
        topic = (
            self.ACTUATOR_STATUS
            + self.DEVICE_PATH_PREFIX
            + device_key
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + actuator_status.reference
        )

        if isinstance(actuator_status.value, bool):
            actuator_status.value = str(actuator_status.value).lower()

        payload = json.dumps(
            {
                "status": actuator_status.state.value,
                "value": str(actuator_status.value),
            }
        )

        message = Message(topic, payload)
        self.log.debug(
            f"Made {message} from {actuator_status} and {device_key}"
        )

        return message

    def make_configuration_message(
        self, device_key: str, configuration: dict
    ) -> Message:
        """
        Make message from configuration for device key.

        :param device_key: Device to which the configuration belongs to.
        :type device_key: str
        :param configuration: Current configuration data
        :type configuration: dict

        :returns: message
        :rtype: Message
        """
        topic = (
            self.CONFIGURATION_STATUS + self.DEVICE_PATH_PREFIX + device_key
        )

        for reference, config_value in configuration.items():
            if isinstance(config_value, tuple):
                values_list = []
                for value in config_value:
                    if "\n" in str(value):
                        value = value.replace("\n", "\\n")
                        value = value.replace("\r", "")

                    if '"' in str(value):
                        value = value.replace('"', '\\"')

                    values_list.append(value)

                configuration[reference] = ",".join(map(str, values_list))
            else:
                if isinstance(config_value, bool):
                    config_value = str(config_value).lower()

                configuration[reference] = str(config_value)

        payload = json.dumps({"values": configuration})

        message = Message(topic, payload)
        self.log.debug(f"Made {message} from {configuration} and {device_key}")

        return message



"""Run this module to install package on your machine."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from os import path

import setuptools

from wolk_gateway_module import __version__

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="wolk-gateway-module",
    version=__version__,
    install_requires=["paho_mqtt>=1.4.0"],
    include_package_data=True,
    license="Apache License 2.0",
    author="WolkAbout",
    author_email="info@wolkabout.com",
    description=(
        "SDK for gateway communication modules "
        "that connect to WolkAbout IoT Platform"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["IoT", "WolkAbout", "Internet of Things"],
    url="https://github.com/Wolkabout/WolkGatewayModule-SDK-Python",
    packages=setuptools.find_packages(exclude=("test",)),
    test_suite="unittest",
    python_requires=">=3.7.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: Communications",
        "Topic :: Software Development :: Embedded Systems",
    ],
)



"""Store data before publishing."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from collections import deque
from typing import List
from typing import Optional

from wolk_gateway_module.logger_factory import logger_factory
from wolk_gateway_module.model.message import Message
from wolk_gateway_module.persistence.outbound_message_queue import (
    OutboundMessageQueue,
)


class OutboundMessageDeque(OutboundMessageQueue):
    """
    Responsible for storing messages before being sent to WolkGateway.

    Messages are sent in the order they were added to the queue.

    Storing readings and alarms without Unix timestamp will result
    in all sent messages being treated as live readings and
    will be assigned a timestamp upon reception, so for a valid
    history add timestamps to readings via `int(round(time.time() * 1000))`
    """

    def __init__(self) -> None:
        """Initialize a double ended queue for storing messages."""
        self.queue: deque = deque()
        self.log = logger_factory.get_logger(str(self.__class__.__name__))

    def __repr__(self) -> str:
        """
        Make string representation of OutboundMessageDeque.

        :returns: representation
        :rtype: str
        """
        return "OutboundMessageDeque()"

    def put(self, message: Message) -> bool:
        """
        Place a message in storage.

        :param message: Message to be stored
        :type message: Message

        :returns: result
        :rtype: bool
        """
        self.log.debug(f"Placing in storage: {message}")
        self.queue.append(message)
        return True

    def remove(self, message: Message) -> bool:
        """
        Remove specific message from storage.

        :returns: result
        :rtype: bool
        """
        self.log.debug(f"Removing from storage: {message}")
        if message in self.queue:
            self.queue.remove(message)
            return True
        return True

    def get_messages_for_device(self, device_key: str) -> List[Message]:
        """
        Return a list of messages that belong to a certain device.

        Does not remove from storage.

        :param device_key: Device identifier
        :type device_key: str

        :returns: messages
        :rtype: List[Message]
        """
        self.log.debug(f"Getting messages for device: {device_key}")
        if self.queue_size() == 0:
            self.log.debug("No messages in queue")
            return []
        messages = []
        for message in self.queue:
            if device_key in message.topic:
                messages.append(message)
        self.log.debug(f"Found messages: {messages}")
        return messages

    def get(self) -> Optional[Message]:
        """
        Get the first message from storage without removing it.

        :returns: message
        :rtype: Message, None
        """
        try:
            message = self.queue[0]
        except IndexError:
            message = None
        self.log.debug(f"Got message from storage: {message}")
        return message

    def queue_size(self) -> int:
        """
        Return current number of messages in storage.

        :returns: size
        :rtype: int
        """
        size = len(self.queue)
        self.log.debug(f"Queue size: {size}")
        return size



"""Device registration response results."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from enum import Enum
from enum import unique


@unique
class DeviceRegistrationResponseResult(Enum):
    """
    Enumeration of possible registration response results.

    :ivar OK: Device was successfully registered
    :vartype OK: str
    :ivar ERROR_GATEWAY_NOT_FOUND: Gateway that sent the registration request was not found
    :vartype ERROR_GATEWAY_NOT_FOUND: str
    :ivar ERROR_NOT_A_GATEWAY: Sender of request is not a gateway
    :vartype ERROR_NOT_A_GATEWAY: str
    :ivar ERROR_KEY_CONFLICT: Device with that key already exists
    :vartype ERROR_KEY_CONFLICT: str
    :ivar ERROR_MAXIMUM_NUMBER_OF_DEVICES_EXCEEDED: Reached limit for number of devices
    :vartype ERROR_MAXIMUM_NUMBER_OF_DEVICES_EXCEEDED: str
    :ivar ERROR_VALIDATION_ERROR: Some data in the registration request was not valid
    :vartype ERROR_VALIDATION_ERROR: str
    :ivar ERROR_INVALID_DTO: The request was not valid - faulty JSON
    :vartype ERROR_INVALID_DTO: str
    :ivar ERROR_KEY_MISSING: Device key was not provided
    :vartype ERROR_KEY_MISSING: str
    :ivar ERROR_SUBDEVICE_MANAGEMENT_FORBIDDEN: Gateway is not able to register devices
    :vartype ERROR_SUBDEVICE_MANAGEMENT_FORBIDDEN: str
    :ivar ERROR_UNKNOWN: Unknown error occurred
    :vartype ERROR_UNKNOWN: str
    """

    OK = "OK"
    ERROR_GATEWAY_NOT_FOUND = "ERROR_GATEWAY_NOT_FOUND"
    ERROR_NOT_A_GATEWAY = "ERROR_NOT_A_GATEWAY"
    ERROR_KEY_CONFLICT = "ERROR_KEY_CONFLICT"
    ERROR_MAXIMUM_NUMBER_OF_DEVICES_EXCEEDED = (
        "ERROR_MAXIMUM_NUMBER_OF_DEVICES_EXCEEDED"
    )
    ERROR_VALIDATION_ERROR = "ERROR_VALIDATION_ERROR"
    ERROR_INVALID_DTO = "ERROR_INVALID_DTO"
    ERROR_KEY_MISSING = "ERROR_KEY_MISSING"
    ERROR_SUBDEVICE_MANAGEMENT_FORBIDDEN = (
        "ERROR_SUBDEVICE_MANAGEMENT_FORBIDDEN"
    )
    ERROR_UNKNOWN = "ERROR_UNKNOWN"



"""Actuator templates used for registering device on WolkAbout IoT Platform."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Dict
from typing import Optional
from typing import Union

from wolk_gateway_module.model.data_type import DataType


class ActuatorTemplate:
    """
    Actuator template for registering device on Platform.

    :ivar description: Description detailing this actuator
    :vartype description: str or None
    :ivar maximum: Maximum actuator value
    :vartype maximum: int or float or None
    :ivar minimum: Minimum actuator value
    :vartype minimum: int of float or None
    :ivar name: Name of actuator
    :vartype name: str
    :ivar reference: Unique actuator reference
    :vartype reference: str
    :ivar unit: Actuator reading type measurement name and unit
    :vartype unit: dict
    """

    def __init__(
        self,
        name: str,
        reference: str,
        data_type: DataType = None,
        reading_type_name: str = None,
        unit: str = None,
        description: str = None,
        minimum: Optional[Union[int, float]] = None,
        maximum: Optional[Union[int, float]] = None,
    ):
        """
        Actuator template for device registration request.

        Define a reading type either by using the ``data_type`` to select a
        generic type (boolean, numeric, string) or use ``reading_type_name``
        and ``unit`` to use a custom reading type that was previously defined
        on WolkAbout IoT Platform.

        :param name: Actuator name
        :type name: str
        :param reference: Actuator reference
        :type reference: str
        :param data_type: Actuator data type
        :type data_type: Optional[DataType]
        :param reading_type_name: Custom reading type name
        :type reading_type_name: Optional[str]
        :param unit: Custom reading type measurement unit
        :type unit: Optional[str]
        :param description: Description detailing the actuator
        :type description: Optional[str]
        :param minimum: Actuator minimum value
        :type minimum: Optional[Union[int,float]]
        :param maximum: Actuator maximum value
        :type maximum: Optional[Union[int,float]]
        """
        self.name: str = name
        self.reference: str = reference
        self.description: Optional[str] = description
        self.minimum: Optional[Union[int, float]] = minimum
        self.maximum: Optional[Union[int, float]] = maximum

        if not (data_type or reading_type_name or unit):
            raise ValueError("Unable to create template")

        if data_type:
            if not isinstance(data_type, DataType):
                raise ValueError("Invalid data type given")
            if data_type == DataType.NUMERIC:
                self.unit: Dict[str, str] = {
                    "readingTypeName": "COUNT(ACTUATOR)",
                    "symbol": "count",
                }
            elif data_type == DataType.BOOLEAN:
                self.unit = {
                    "readingTypeName": "SWITCH(ACTUATOR)",
                    "symbol": "",
                }
            elif data_type == DataType.STRING:
                self.unit = {"readingTypeName": "TEXT(ACTUATOR)", "symbol": ""}
            return

        if not (reading_type_name and unit):
            raise ValueError(
                "Both reading type name and unit must be provided"
            )
        self.unit = {"readingTypeName": reading_type_name, "symbol": unit}

    def __repr__(self) -> str:
        """
        Make string representation of actuator template.

        :returns: representation
        :rtype: str
        """
        return (
            f"ActuatorTemplate(name='{self.name}', "
            f"reference='{self.reference}', description='{self.description}', "
            f"unit='{self.unit}', "
            f"minimum='{self.minimum}', maximum='{self.maximum}')"
        )

    def to_dto(self) -> Dict[str, Union[int, float, str, Dict[str, str]]]:
        """
        Create data transfer object used for registration.

        :returns: dto
        :rtype: Dict[str, Union[int, float, str]]
        """
        dto: Dict[str, Union[int, float, str, Dict[str, str]]] = {
            "name": self.name,
            "reference": self.reference,
            "unit": self.unit,
        }

        dto["description"] = self.description if self.description else ""

        if self.minimum is not None and self.maximum is not None:
            dto["minimum"] = self.minimum
            dto["maximum"] = self.maximum

        return dto



"""Alarm event model."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from typing import Optional


@dataclass
class Alarm:
    """
    Holds information about a devices alarm.

    :ivar reference: Device alarm's reference as defined in device template
    :vartype reference: str
    :ivar active: Alarm's current state
    :vartype active: bool
    :ivar timestamp: Unix timestamp in miliseconds. If not provided, Platform will assign timestamp when it receives it.
    :vartype timestamp: int or None
    """

    reference: str
    active: bool
    timestamp: Optional[int]



"""Tests for MQTTConnectivityService."""
#   Copyright 2019 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import sys
import unittest

sys.path.append("..")  # noqa

from wolk_gateway_module.mqtt_connectivity_service import (
    MQTTConnectivityService,
)
from wolk_gateway_module.model.message import Message


class MQTTConnectivityServiceTests(unittest.TestCase):
    """MQTTConnectivityService Tests."""

    def mock_listener(self, message):
        """Mock function for inbound message listener."""
        self.message = message

    def set_up(self):
        """Prepare context for testing."""
        self.host = "localhost"
        self.port = 1883
        self.client_id = "WolkGatewayModule-SDK-Python"
        self.qos = 0
        self.lastwill_message = Message("lastwill")
        self.topics = []
        self.listener = self.mock_listener
        self.message = None
        self.mqttcs = MQTTConnectivityService(
            self.host,
            self.port,
            self.client_id,
            self.qos,
            self.lastwill_message,
            self.topics,
        )
        self.mqttcs.log.setLevel(logging.CRITICAL)

    def tear_down(self):
        """Clean context from completed test."""
        self.host = None
        self.port = None
        self.client_id = None
        self.qos = None
        self.lastwill_message = None
        self.topics = None
        self.listener = None
        self.mqttcs = None
        self.message = None

    def test_set_inbound_message_listener(self):
        """Test that inbound message listener is set correctly."""
        self.set_up()

        self.mqttcs.set_inbound_message_listener(self.mock_listener)

        self.assertEqual(
            self.mock_listener, self.mqttcs.inbound_message_listener
        )

        self.tear_down()

    def test_set_lastwill_message(self):
        """Test that lastwill message is set correctly."""
        self.set_up()

        message = Message("lastwill")

        self.mqttcs.set_lastwill_message(message)

        self.assertEqual(message, self.mqttcs.lastwill_message)

        self.tear_down()

    def test_add_subscription_topics(self):
        """Test adding subscription topics."""
        self.set_up()

        topics = ["topic1", "topic2", "topic3"]

        self.mqttcs.add_subscription_topics(topics)

        self.assertEqual(3, len(self.mqttcs.topics))

        self.tear_down()

    def test_remove_topics_for_device(self):
        """Test removing topics from subscription topics."""
        self.set_up()

        self.mqttcs.topics = ["key1", "key2", "key3", "key1"]

        device_key = "key1"

        self.mqttcs.remove_topics_for_device(device_key)

        self.assertEqual(2, len(self.mqttcs.topics))

        self.tear_down()

    def test_inbound_message_listener(self):
        """Test that inbound message listener is called correctly."""
        self.set_up()

        self.mqttcs.set_inbound_message_listener(self.mock_listener)

        expected = Message("test_message")

        self.mqttcs._on_mqtt_message(None, None, expected)

        self.assertEqual(expected, self.message)

        self.tear_down()

    def test_connect(self):
        """Test that connect is called correctly."""
        self.set_up()

        self.mqttcs.connected_rc = 0

        self.assertTrue(self.mqttcs.connect())

        self.tear_down()

    def test_connected(self):
        """Test connected returns false."""
        self.set_up()

        self.assertFalse(self.mqttcs.connected())

        self.tear_down()

    def test_connect_raises_exception(self):
        """Test that connect raises exception."""
        self.set_up()

        self.mqttcs.connected_rc = 5

        try:
            self.mqttcs.connect()
        except Exception as expection:
            self.assertTrue(isinstance(expection, RuntimeError))

        self.tear_down()

    def test_reconnect(self):
        """Test that reconnection method will call connect."""
        self.set_up()

        self.mqttcs.connected_rc = 0

        self.assertTrue(self.mqttcs.reconnect())

        self.tear_down()

    def test_publish_fails(self):
        """Test that publishing fails when not connected."""
        self.set_up()

        message = Message("test")

        self.assertFalse(self.mqttcs.publish(message))

        self.tear_down()


if __name__ == "__main__":
    unittest.main()
